/*
 * File:         ers7_jen.c
 * Date:         Januar 12 , 2007
 * Description:  A controller for the real Aibo ERS-7 robot.
 * Author:       Jennifer Meinen
 */

#include <device/robot.h>
#include <device/servo.h>
#include <device/camera.h>
#include <math.h>
#include <time.h>


//Control step duration in milliseconds for simulation
#define SIMULATION_STEP 32

//Specific to the robot from Model Information for ERS-7
#define FOCAL_LENGTH 3.27 // mm
#define HORIZ_ANGLE_VIEW 56.9 //degree
#define VERT_ANGLE_VIEW 45.2 //degree
#define NECK 80
#define HEAD_Y 81.06
#define HEAD_Z 14.6

//Matrices maximum size
#define N 4 
typedef struct {
   double numbers[N][N];
   unsigned int lignes;
   unsigned int colonnes;
} Matrice;


/* prototypes */
void  affiche_matrice(Matrice);
Matrice mult(Matrice a, Matrice b);
Matrice inverse3x3(Matrice a); 
Matrice convert_refBody_to_refCam(Matrice Pbody);
Matrice convert_refCam_to_refBody(Matrice Pcam);
int intersect_plane_line(Matrice * point_result,Matrice La,Matrice Lb,Matrice P0,Matrice P1,Matrice P2);
Matrice convert_inhomogenous(Matrice M);
double convert_radian(double degre);
void calcul_milieu_parallelograme(double A[2],double B[2],double C[2],double D[2],double result[2]);

int get_mark_pixel(double pixel[2]);
int calculate_pixel_distance(double pixelPosition[2],double pixelDistance[3]);
int isRed_or_Magenta(int r,int g,int b);
void set_initial_position(double j1_left_f,double j3_left_f,double j1_right_f,double j3_right_f,double j1_left_b,double j3_left_b,double j1_right_b,double j3_right_b);
void calculate_P0_P1_P2(Matrice * P0,Matrice * P1,Matrice * P2);
int check_aibobody_correct_position();
void calculate_mark_coordinate();
int seek(int found, double pixel[2]);
int get_first_mark_pixel(double pixel[2]);
void move_nearest_paw(Matrice mark);

Matrice calculate_angles_cin(double x, double y,double z);
Matrice tilt1_to_front_leg_right(double x,double y,double z);
Matrice tilt1_to_front_leg_left(double x,double y,double z);
void move_paw(double x_scene_tree,double y_scene_tree,double z_scene_tree);



/*variable globales*/
static double alpha=0; //rotation tilt1
static double betha=0; //rotation pan
static double omega=0; //rotation tilt2
static int only_once=0;//experience is made only once
static double Color_tolerance = 0.1; //color tolerance
static int pan_moving_left=0;//head is moving left if pan_moving_left is 1
static int mark_found=0;//


//sert  dfinir la position initiale dans lequel on calcul P0, P1, P2
static double angle_left_front_J1, angle_left_front_J3, angle_right_front_J1, angle_right_front_J3, angle_left_back_J1, angle_left_back_J3, angle_right_back_J1, angle_right_back_J3;

//declarer tous les devices
static DeviceTag camera, neck_tilt1, neck_tilt2, neck_pan, Left_front_legJ1, Left_front_legJ2, Left_front_legJ3, Left_rear_legJ1, Left_rear_legJ2, Left_rear_legJ3, Right_front_legJ1 ;
static DeviceTag Right_front_legJ2, Right_front_legJ3, Right_rear_legJ1, Right_rear_legJ2, Right_rear_legJ3;



static void init(void)
{
  /* enable camera */
  camera = robot_get_device("PRM:/r1/c1/c2/c3/i1-FbkImageSensor:F1");
  camera_enable(camera, SIMULATION_STEP);
 
  /* head */
  neck_tilt1 = robot_get_device("PRM:/r1/c1-Joint2:11");
  neck_pan = robot_get_device("PRM:/r1/c1/c2-Joint2:12");
  neck_tilt2 = robot_get_device("PRM:/r1/c1/c2/c3-Joint2:13");
  
  servo_enable_position  (neck_tilt2, SIMULATION_STEP);
  omega=0.05;
  servo_set_position (neck_tilt2,omega);
  
  servo_enable_position  (neck_pan, SIMULATION_STEP);
  betha=0;
  servo_set_position (neck_pan,betha);
  
  servo_enable_position  (neck_tilt1, SIMULATION_STEP);
  alpha=-1.3;
  servo_set_position (neck_tilt1,alpha);
  
  /* body */
  Left_front_legJ1 = robot_get_device("PRM:/r2/c1-Joint2:21");
  Left_front_legJ2 = robot_get_device("PRM:/r2/c1/c2-Joint2:22");
  Left_front_legJ3 = robot_get_device("PRM:/r2/c1/c2/c3-Joint2:23");
  
  Left_rear_legJ1 = robot_get_device("PRM:/r3/c1-Joint2:31");
  Left_rear_legJ2 = robot_get_device("PRM:/r3/c1/c2-Joint2:32");
  Left_rear_legJ3 = robot_get_device("PRM:/r3/c1/c2/c3-Joint2:33");
  
  Right_front_legJ1 = robot_get_device("PRM:/r4/c1-Joint2:41");
  Right_front_legJ2 = robot_get_device("PRM:/r4/c1/c2-Joint2:42");
  Right_front_legJ3 = robot_get_device("PRM:/r4/c1/c2/c3-Joint2:43");
  
  Right_rear_legJ1 = robot_get_device("PRM:/r5/c1-Joint2:51");
  Right_rear_legJ2 = robot_get_device("PRM:/r5/c1/c2-Joint2:52");
  Right_rear_legJ3 = robot_get_device("PRM:/r5/c1/c2/c3-Joint2:53");
  servo_enable_position  (Left_front_legJ1, SIMULATION_STEP);
  servo_enable_position  (Left_front_legJ2, SIMULATION_STEP);
  servo_enable_position  (Left_front_legJ3, SIMULATION_STEP);
  
  servo_enable_position  (Left_rear_legJ1, SIMULATION_STEP);
  servo_enable_position  (Left_rear_legJ2, SIMULATION_STEP);
  servo_enable_position  (Left_rear_legJ3, SIMULATION_STEP);
  
  servo_enable_position  (Right_front_legJ1, SIMULATION_STEP);
  servo_enable_position  (Right_front_legJ2, SIMULATION_STEP);
  servo_enable_position  (Right_front_legJ3, SIMULATION_STEP);
  
  servo_enable_position  (Right_rear_legJ1, SIMULATION_STEP);
  servo_enable_position  (Right_rear_legJ2, SIMULATION_STEP);
  servo_enable_position  (Right_rear_legJ3, SIMULATION_STEP);
  
  angle_left_front_J1=convert_radian(0);
  angle_left_front_J3=convert_radian(30);
  
  angle_right_front_J1=convert_radian(0);
  angle_right_front_J3=convert_radian(30);
  
  angle_left_back_J1=convert_radian(0);
  angle_left_back_J3=convert_radian(30);
  
  angle_right_back_J1=convert_radian(0);
  angle_right_back_J3=convert_radian(30);
  
  set_initial_position(angle_left_front_J1,angle_left_front_J3,angle_right_front_J1,angle_right_front_J3,angle_left_back_J1,angle_left_back_J3,angle_right_back_J1,angle_right_back_J3);
  
    return;
}

static int run(int ms)
{
    time_t start,end;
    if(!mark_found){
    double pixel[2];
    mark_found=seek(mark_found,pixel);
    time (&start);
    }
    
    if(mark_found && only_once < 1){  //&& check_aibobody_correct_position()
      //to be sure Aibo will not move before calculating we start a counter
	    time (&end);
	    if(difftime (end,start) > 5 )
	    {
        printf("Mark found\n");
        
	      angle_left_front_J1=servo_get_position(Left_front_legJ1); 
      	angle_left_front_J3=servo_get_position(Left_front_legJ3); 
	      angle_right_front_J1=servo_get_position(Right_front_legJ1);
      	angle_right_front_J3=servo_get_position(Right_front_legJ3); 
	      angle_left_back_J1=servo_get_position(Left_rear_legJ1); 
	      angle_left_back_J3=servo_get_position(Left_rear_legJ3);
      	angle_right_back_J1=servo_get_position(Right_rear_legJ1); 
      	angle_right_back_J3=servo_get_position(Right_rear_legJ3);
        //to test without
	      //set_initial_position(angle_left_front_J1, angle_left_front_J3, angle_right_front_J1, angle_right_front_J3, angle_left_back_J1, angle_left_back_J3,angle_right_back_J1,angle_right_back_J3);
        
	      //Pareil pour la tete
      	omega=servo_get_position(neck_tilt2);
      	betha=servo_get_position(neck_pan);
      	alpha=servo_get_position(neck_tilt1);
	      servo_set_position (neck_tilt2,omega);
  	    servo_set_position (neck_pan,betha);
  	    servo_set_position (neck_tilt1,alpha);

        printf("neck_tilt2 : %f\n",omega);
        printf("neck_pan : %f\n",betha);
        printf("neck_tilt1 : %f\n",alpha);
       	calculate_mark_coordinate();
	      only_once++;
	  }else{
	  printf("i m waiting");
	  }
  }
    
  return SIMULATION_STEP;
}

static void die(void)
{
    return;
}

int main(void)
{
    robot_live(init);
    robot_die(die);
    robot_run(run);

    return 0;
}

/* ====================================================================== */
void calculate_mark_coordinate(){
printf("------------Start\n");
      Matrice O = {   {
                          {0},
                          {0},
                          {0} } , 3 , 1 };
      double pixelPosition[2];
      double pixelDistance[3];
      Matrice G1,G2, G3;
      calculate_P0_P1_P2(&G1,&G2,&G3);
      
      
          
      if(get_mark_pixel(pixelPosition)){
         
        Matrice G1_cam = convert_refBody_to_refCam(G1);
        Matrice G2_cam = convert_refBody_to_refCam(G2);
        Matrice G3_cam = convert_refBody_to_refCam(G3);
        
        printf("G1 dans referentiel de la camera : ");
        affiche_matrice (G1_cam);
        printf("\n");
        
        calculate_pixel_distance(pixelPosition,pixelDistance);
        
        Matrice Lb = {   {
                                      {pixelDistance[0]},
                                      {pixelDistance[1]},
                                      {pixelDistance[2]} } , 3 , 1 };
        printf("Reprsente la distance reel d'un pixel : ");
        affiche_matrice (Lb);
        printf("\n");

        Matrice point_result;
        
        if( intersect_plane_line(&point_result,O,Lb,convert_inhomogenous(G1_cam),convert_inhomogenous(G2_cam),convert_inhomogenous(G3_cam))  )
        {
            //un point a t trouv : point_result est homogne on peut donc appliquer directement la conversion dans l'autre referentiel et affich le rsultat dans la console
            printf("Point dans le referentiel de la camera : ");
            affiche_matrice (point_result);
            printf("\n");
            Matrice result_final=convert_refCam_to_refBody(point_result);

            printf("Point dans le referentiel de tilt1 : ");
            affiche_matrice (result_final);
            printf("\n");
            printf("Position calculee dans scene tree: x= %.4g , y= %.4g , z= %.4g",result_final.numbers[0][0]/1000,result_final.numbers[2][0]/1000,-result_final.numbers[1][0]/1000);
  
            //conversion de referentiel x=point_result.numbers[0][0]/1000   y=point_result.numbers[0][2]/1000  z=-point_result[0][1]/1000
            move_paw(result_final.numbers[0][0]/1000,result_final.numbers[2][0]/1000,-result_final.numbers[1][0]/1000);
            //move_nearest_paw(result_final)
            
        }
        else
        {
        printf("Aucune intersection avec le plan n a ete trouvee");
        }
      }
      else
      {
        printf("Pas de pixel rouge");
      }
	printf("---------End\n");

}
/* ====================================================================== */
int mark_middle(double pixelPosition[2]){

 float h = camera_get_width  (camera) /2.0;
 float x=pixelPosition[0]-h;
 
 if(fabs(x)<10) 
 {
  return 1;
 }else{
  if(pan_moving_left){
    betha+=0.1;
  }else{
    betha-=0.1;
  }
  servo_set_position(neck_pan,betha);
  return 0;
 }
}

int hardware_bounds_reached(double a){
  if(betha>1.3 || betha<-1.3){
    return 1;
  }else{
    return 0;
  }
}
int seek(int found, double pixel[2]){
//found = 1 si marque deja trouvee

if(!found){
  if(get_first_mark_pixel(pixel)){
  //ici on optimise et deplacer la tete jusqua ce que la marque soit le plus au centre
  //mettre alpha betha omega a cette position donc plus de mouvement
  
  if(mark_middle(pixel) || hardware_bounds_reached(betha)){
  
    omega=servo_get_position (neck_tilt2);
    betha=servo_get_position  (neck_pan);
    alpha=servo_get_position (neck_tilt1);
    
    servo_set_position (neck_tilt2,omega);
    servo_set_position (neck_pan,betha);
    servo_set_position (neck_tilt1,alpha); //tete ne bouge plus
    return 1;
  }
  else
  {
    return 0;
  }
    
  }else{
  //tourner la tete a gauche jusqua la limite
  //si la tete a deja atteint l obectif fixe par alpha regarder si on est en train de tourner a gauche ou a droite et si on a pas atteint la limite on incremente, sinon on repart dans l autre sens
 
  if(fabs(servo_get_position(neck_pan)-betha)<0.1)
	{

		if(pan_moving_left)
		{
      			if(betha<0.6)
			{
        			betha+=0.1;
        			servo_set_position(neck_pan,betha);
      			}
			else
			{
      			//limite atteinte on repart dans l autre sens
      				pan_moving_left=0;
				printf("Changement de sens tete 1\n");
      			}
    		}
		else
		{
    		//on est en train de tourner a droite
    			if(betha>-0.6)
			{
        			betha-=0.1;
        			servo_set_position(neck_pan,betha);
      			}
			else
			{
      			//limite atteinte on repart dans l autre sens
      			pan_moving_left=1;
			printf("Changement de sens tete 2\n");
      			}
    		}
  	}else{
	//printf("Position actuel de la tete %f position voulue %f\n",servo_get_position(neck_pan),betha);
	}
  
  return 0;
  }
}
else
{
  return 1;
}

}

/* ====================================================================== */

//tenant compte de angle de vue, focale en considrant un modle pinhole camera
//retourne 0 si tout se passe bien et 1 si il y a un problme
int calculate_pixel_distance(double pixelPosition[2],double pixelDistance[3]){
 
 
 float image_width = camera_get_width  (camera);
 float image_height = camera_get_height (camera); 
 
 float h = image_width/2;
 float v = image_height/2;
 double a = convert_radian(HORIZ_ANGLE_VIEW/2.0);
 double b = convert_radian(VERT_ANGLE_VIEW/2.0); //TESTER
  
 printf("Pixel depuis le bord (%f,%f)\n",pixelPosition[0],pixelPosition[1]);
 printf("Pixel dans le referentiel x,y milieu de l'image (%f,%f)\n",pixelPosition[0]-h,v-pixelPosition[1]);
 
 pixelDistance[0]= ((pixelPosition[0]-h)* FOCAL_LENGTH * tan(a))/h;
 pixelDistance[1]= FOCAL_LENGTH;
 pixelDistance[2]= ((v-pixelPosition[1])* FOCAL_LENGTH * tan(b))/v;
 
 return 0;
}

/* ====================================================================== */
//retourne la position du premier pixel qu'il voit
int get_first_mark_pixel(double pixel[2]){

  const unsigned char *image = camera_get_image(camera);
  unsigned short image_width = camera_get_width  (camera);
  unsigned short image_height = camera_get_height (camera); 
  //printf("image_width=%d, image_height=%d", image_width, image_height);//voir ou s'affiche ce print
  int x;
  int y;
  int nbr_pixel=0;
  int nbr_pixel_line=0;
  int x_temp;
  int y_temp;
 
 for (y = 0; y < image_height; y++){ //ligne de l'image
    for (x = 0; x < image_width; x++) { //colonne de l'image
      int r = camera_image_get_red(image, image_width, x, y);
      int g = camera_image_get_green(image, image_width, x, y);
      int b = camera_image_get_blue(image, image_width, x, y);
      //printf("red=%d, green=%d, blue=%d", r, g, b);//voir ou s'affiche ce print
      if(isRed_or_Magenta(r,g,b)){
        x_temp=x;
        y_temp=y;
        nbr_pixel++;
	nbr_pixel_line++;
      }
    }
    if(nbr_pixel_line>2){
      pixel[0]=x_temp;
      pixel[1]=y_temp;
    }
    nbr_pixel_line=0;

  }
    
  if(nbr_pixel<5){
    return 0;
  }else{
    printf("TUT \n");
    return 1;
  }
}
  
/* ====================================================================== */
//prend une image trouve si il y a un certain niveau de rouge plus lev et donne la position du pixel du milieu
//retourne 1 si il a trouver un pixel rouge/magenta
int get_mark_pixel(double pixel[2]){

  const unsigned char *image = camera_get_image(camera);
  unsigned short image_width = camera_get_width  (camera);
  unsigned short image_height = camera_get_height (camera); 
  //printf("image_width=%d, image_height=%d", image_width, image_height);//voir ou s'affiche ce print
  int x;
  int y;
  
  double A[2];
  double B[2];
  double C[2];
  double D[2];
  
  double A_temp[2];
  double B_temp[2];
  
  
  int nb_red_ligne_prec=0;
  int nb_red_ligne_curr=0;
  int has_red=0;


 for (y = 0; y < image_height; y++){ //ligne de l'image
    for (x = 0; x < image_width; x++) { //colonne de l'image
      int r = camera_image_get_red(image, image_width, x, y);
      int g = camera_image_get_green(image, image_width, x, y);
      int b = camera_image_get_blue(image, image_width, x, y);
      //printf("red=%d, green=%d, blue=%d", r, g, b);//voir ou s'affiche ce print
      
      
      if(isRed_or_Magenta(r,g,b)){
        if(nb_red_ligne_prec==0){
          //on vient d'arriver dans la ligne de notre marque
	   printf("red=%d, green=%d, blue=%d", r, g, b);
	   printf("Position du premier pixel %d %d",x,y);
          if(nb_red_ligne_curr==0){
            //c'est le tout premier pixel de la ligne
            has_red=1;
            A[0]=x;
            A[1]=y;
            C[0]=x;
            C[1]=y;
            B[0]=x;
            B[1]=y;
            D[0]=x;
            D[1]=y;
            nb_red_ligne_curr=1;
          }else{
            //ce n'est pas le premier pixel de la ligne
            B[0]=x;
            B[1]=y;
            D[0]=x;
            D[1]=y;
            nb_red_ligne_curr++;
          }
          
        }else{
          //la ligne precedente contenait deja des pixel rouge
          if(nb_red_ligne_curr==0){
              //c'est le tout premier pixel de la ligne , la ligne prcdente contenait dj des pixel rouge
              //A doit changer seulement si le nb de pixel de la ligne prcdente est strictement plus petit que cette ligne ci
              A_temp[0]=x;
              A_temp[1]=y;
              B_temp[0]=x;
              B_temp[1]=y;
              nb_red_ligne_curr=1;
            }else{
              //ce n'est pas le premier pixel de la ligne
              B_temp[0]=x;
              B_temp[1]=y;
              nb_red_ligne_curr++;
            }
        }
      }
      
    }
    
    if(!(nb_red_ligne_prec==0) && nb_red_ligne_prec<nb_red_ligne_curr){ 
      //A et B doivent changer
      A[0]=A_temp[0];
      A[1]=A_temp[1];
      B[0]=B_temp[0];
      B[1]=B_temp[1];
    }
    
    if(!(nb_red_ligne_prec==0) && nb_red_ligne_prec==nb_red_ligne_curr){ 
      //A et B doivent changer
      C[0]=A_temp[0];
      C[1]=A_temp[1];
      D[0]=B_temp[0];
      D[1]=B_temp[1];
    }
    //on a parcouru tous les pixels d'une ligne
    if(nb_red_ligne_curr !=0){
      printf("Nombre de pixel rouge %d\n",nb_red_ligne_curr);
    }
    nb_red_ligne_prec=nb_red_ligne_curr;
    nb_red_ligne_curr=0;
}



if(has_red){
  calcul_milieu_parallelograme(A,B,C,D,pixel);
  return 1;
}else{
  return 0;
}


}

/* ====================================================================== */

//retourne 1 si le pixel est rouge/magenta
int isRed_or_Magenta(int r,int g,int b){
//printf("isRed_or_Magenta red=%d, green=%d, blue=%d", r, g, b);//voir ou s'affiche ce print
  if(  r>(255.0-(Color_tolerance*255.0)) && g<100 ){  //&& g<Color_tolerance*255
    return 1;
  }else{
    return 0;
  }
}

//retourne 1 si le pixel est rouge not used yet ->perhaps for later
/*
int isRed(int r,int g,int b){
  if(r>(255-Color_tolerance*255) && g<Color_tolerance*255 && b<Color_tolerance*255){
    return 1;
  }else{
    return 0;
  }
}*/


/* ====================================================================== */
Matrice mult(Matrice a, Matrice b)
{
  Matrice resultat;
  unsigned int i, j, k;

  resultat.lignes = a.lignes;
  resultat.colonnes = b.colonnes;
  
  if (a.colonnes == b.lignes) {
    
    for (i = 0; i < a.lignes; ++i) {       /* Pour chaque ligne de la matrice a */
      for (j = 0; j < b.colonnes; ++j) {   /* Pour chaque colonne de la matrice b */
        resultat.numbers[i][j] = 0.0;
	  for (k = 0; k < b.lignes; ++k) { /* Pour chaque lment de la ligne i de a et la colonne j de b */
	  resultat.numbers[i][j] += a.numbers[i][k] * b.numbers[k][j];
        }
      }
    }
  }

  return resultat;
}


/* ====================================================================== */
void affiche_matrice(Matrice m)
{
  unsigned int i, j;
  for (i = 0; i < m.lignes; ++i) {
    for (j = 0; j < m.colonnes; ++j) 
      printf("%g ", m.numbers[i][j]);
    putchar('\n');
  }
}



/* ====================================================================== */
Matrice inverse3x3(Matrice A)
{
  Matrice resultat;
  resultat.lignes=0; //si elle ne peut pas etre remplie la taille reste 0
  resultat.colonnes=0;
  /*
	  a   b  c
M  =     d   e  f
	  g   h  i    
	  
det(M)=  a(ei-fh) - b(di-fg) + c(dh-eg) 

			ei-fh	ch-bi	bf-ce
M^-1 =  1/det(m) *	fg-di	ai-cg	cd-af
			dh-eg	bg-ah	ae-bd

  */
  //verifier que la matrice est bien carre et qu'elle est 3x3
  if(A.lignes == A.colonnes && A.lignes == 3){
	  //verifier qu'elle est bien inversible
	  double a=A.numbers[0][0], b=A.numbers[0][1], c=A.numbers[0][2],
			 d=A.numbers[1][0], e=A.numbers[1][1], f=A.numbers[1][2],
			 g=A.numbers[2][0], h=A.numbers[2][1], i=A.numbers[2][2];
	  double detA = a*(e*i-f*h) - b*(d*i-f*g) + c*(d*h-e*g);
	  //printf("%f",detA);
	  if(fabs(detA)>0)//detA doit etre autre que 0
	  {
	  	  resultat.numbers[0][0]=(e*i-f*h)/detA;
		  resultat.numbers[0][1]=(c*h-b*i)/detA;
		  resultat.numbers[0][2]=(b*f-c*e)/detA;
		  
		  resultat.numbers[1][0]=(f*g-d*i)/detA;
		  resultat.numbers[1][1]=(a*i-c*g)/detA;
		  resultat.numbers[1][2]=(c*d-a*f)/detA;
		  
		  resultat.numbers[2][0]=(d*h-e*g)/detA;
		  resultat.numbers[2][1]=(b*g-a*h)/detA;
		  resultat.numbers[2][2]=(a*e-b*d)/detA;
		  
		  resultat.lignes=3; 
		  resultat.colonnes=3;
	  } 
  }
  return resultat;
}

/* ====================================================================== */


Matrice rotateX(double tetha){
	Matrice aroundX = {  {{1,0,0,0},
                        {0,cos(tetha),sin(tetha),0},
                        {0,-sin(tetha),cos(tetha),0},
                        {0,0,0,1} } , 4 , 4 };
	return aroundX;
}

Matrice rotateZ(double tetha){
	Matrice aroundZ = {  {{cos(tetha),sin(tetha),0,0},
                      {-sin(tetha),cos(tetha),0,0},
                      {0,0,1,0},
                      {0,0,0,1} } , 4 , 4 };
	return aroundZ;
}

Matrice translate(double a,double b,double c){
	Matrice translation = {  {{1,0,0,-a},
                        {0,1,0,-b},
                        {0,0,1,-c},
                        {0,0,0,1} } , 4 , 4 };
	return translation;
}

Matrice convert_refBody_to_refCam(Matrice Pbody){
  Matrice aroundX1 = rotateX(alpha);
  Matrice aroundZ = rotateZ(betha);
  Matrice trans1 = translate(0,0,NECK);
  Matrice aroundX2 = rotateX(omega);
  Matrice trans2 = translate(0,HEAD_Y,-HEAD_Z);
  /*
  printf("\n after Rotation around X : ");
  affiche_matrice(mult(aroundX1,Pbody));
  printf("\n after Rotation around Z : ");
  affiche_matrice(mult(aroundZ, mult(aroundX1,Pbody)));
  printf("\n after translation : ");
  affiche_matrice(mult(trans1, mult(aroundZ, mult(aroundX1,Pbody) ) ));
  printf("\n after Rotation around X2 : ");
  affiche_matrice(mult(aroundX2 , mult(trans1, mult(aroundZ, mult(aroundX1,Pbody) ) ) ));
  printf("\n after translation : ");
  affiche_matrice(mult(trans2 , mult(aroundX2 , mult(trans1, mult(aroundZ, mult(aroundX1,Pbody) ) ) ) ));
  */
  return mult(trans2 , mult(aroundX2 , mult(trans1, mult(aroundZ, mult(aroundX1,Pbody) ) ) ) );
}

Matrice convert_refCam_to_refBody(Matrice Pcam){
  Matrice aroundX1 = rotateX(-alpha);
  Matrice aroundZ = rotateZ(-betha);
  Matrice trans1 = translate(0,0,-NECK);
  Matrice aroundX2 = rotateX(-omega);
  Matrice trans2 = translate(0,-HEAD_Y,HEAD_Z);
  /*
  printf("\n after translation : ");
  affiche_matrice(mult(trans2,Pcam));
  printf("\n after Rotation around X2 : ");
  affiche_matrice(mult(aroundX2,mult(trans2,Pcam)));
  printf("\n after translation : ");
  affiche_matrice(mult(trans1,mult(aroundX2,mult(trans2,Pcam))));
  printf("\n after Rotation around Z : ");
  affiche_matrice(mult(aroundZ,mult(trans1,mult(aroundX2,mult(trans2,Pcam)))));
  printf("\n after Rotation around X : ");
  affiche_matrice(mult(aroundX1 , mult(aroundZ,mult(trans1,mult(aroundX2,mult(trans2,Pcam))))));
  */
  return mult(aroundX1 , mult(aroundZ,mult(trans1,mult(aroundX2,mult(trans2,Pcam)))));
}

/* ====================================================================== */

//renvois 1 si un point a t trouv et 0 sinon, stock le rsultat dans point_result
int intersect_plane_line(Matrice * point_result,Matrice La,Matrice Lb,Matrice P0,Matrice P1,Matrice P2){
	Matrice M={
				{
					{La.numbers[0][0]-Lb.numbers[0][0],P1.numbers[0][0]-P0.numbers[0][0],P2.numbers[0][0]-P0.numbers[0][0]},
					{La.numbers[1][0]-Lb.numbers[1][0],P1.numbers[1][0]-P0.numbers[1][0],P2.numbers[1][0]-P0.numbers[1][0]},
					{La.numbers[2][0]-Lb.numbers[2][0],P1.numbers[2][0]-P0.numbers[2][0],P2.numbers[2][0]-P0.numbers[2][0]}
				}, 3, 3 };
	Matrice V={
				{
					{La.numbers[0][0]-P0.numbers[0][0]},
					{La.numbers[1][0]-P0.numbers[1][0]},
					{La.numbers[2][0]-P0.numbers[2][0]}
				}, 3, 1 };
	int i;
	
	
	Matrice Temp=inverse3x3(M);
	
	if(Temp.lignes != 0)
	{
		Temp=mult(Temp,V);
		
		for(i=0;i<3;i++){
		  point_result->numbers[i][0]=La.numbers[i][0]+Temp.numbers[0][0]*(Lb.numbers[i][0]-La.numbers[i][0]);
		}
		point_result->numbers[3][0]=1;
		point_result->lignes=4;
		point_result->colonnes=1;

		return 1;
	}
	else
	{
		return 0;
	}

	return 0;
}


/* ====================================================================== */

Matrice convert_inhomogenous(Matrice M){
Matrice result;
int i;
if(M.lignes == 4 && M.colonnes == 1) {
	for(i=0;i<M.lignes-1;i++){
	result.numbers[i][0]=M.numbers[i][0]/M.numbers[3][0];
	}
	result.lignes = 3;
	result.colonnes = 1;
}else if(M.lignes == 1 && M.colonnes == 4){
	for(i=0;i<M.colonnes-1;i++){
	result.numbers[0][i]=M.numbers[0][i]/M.numbers[0][3];
	}
	result.lignes = 1;
	result.colonnes = 3;
}else {
	 //inconvertible
	 result.lignes = 0;
	 result.colonnes = 0;
}
return result;
}

/* ====================================================================== */

double convert_radian(double degre){
return (degre*2*M_PI)/360.0;
}

/* ====================================================================== */

void calcul_milieu_parallelograme(double A[2],double B[2],double C[2],double D[2],double result[2]){

 double A_car[2];
 double B_car[2];
 double C_car[2];
 
if(A[0]<=C[0])
{
	A_car[0]=C[0];
	C_car[0]=C[0];
}
else
{
	A_car[0]=A[0];
	C_car[0]=A[0];
}
A_car[1]=A[1];
C_car[1]=C[1];



if(B[0]<=D[0])
{
	B_car[0]=B[0];
}
else
{
	B_car[0]=D[0];
}
B_car[1]=B[1];

//printf("Le A_car est (%f,%f)\n",A_car[0],A_car[1]);
//printf("Le B_car est (%f,%f)\n",B_car[0],B_car[1]);
//printf("Le C_car est (%f,%f)\n",C_car[0],C_car[1]);

result[0]=A_car[0]+((B_car[0]-A_car[0])/2);
result[1]=A_car[1]+((C_car[1]-A_car[1])/2);
}

/* ====================================================================== */

void set_initial_position(double j1_left_f,double j3_left_f,double j1_right_f,double j3_right_f,double j1_left_b,double j3_left_b,double j1_right_b,double j3_right_b){

  servo_set_position(Left_front_legJ1, j1_left_f);
  servo_set_position(Left_front_legJ2, 0);
  servo_set_position(Left_front_legJ3, j3_left_f);
  
  servo_set_position(Right_front_legJ1, j1_right_f);
  servo_set_position(Right_front_legJ2, 0);
  servo_set_position(Right_front_legJ3, j3_right_f);
  
  servo_set_position(Left_rear_legJ1, j1_left_b);
  servo_set_position(Left_rear_legJ2, 0);
  servo_set_position(Left_rear_legJ3, j3_left_b);
  
  servo_set_position(Right_rear_legJ1, j1_right_b);
  servo_set_position(Right_rear_legJ2, 0);
  servo_set_position(Right_rear_legJ3, j3_right_b);
}


/* ====================================================================== */
void calculate_P0_P1_P2(Matrice * P0,Matrice * P1,Matrice * P2){
  double const w1=19.5;
  double const w2=2.5;
  double const w3=132.5;
  double const w4=134.4;
  double const L1=70.0803110724;
  double const L2=76.8969440225;
  double const L3=79.4099489989;
  double const alpha_init=atan(9.0/69.5);
  double const teta_init=-alpha_init;
  double const beta_init=convert_radian(30)-atan(28.3/71.5);
  double const sig_init=convert_radian(30)-atan(21.3/76.5);
  
  //patte droite et gauche sont solidaire
  double a=servo_get_position(Left_front_legJ1); 
  double b=servo_get_position(Left_front_legJ3);
      
  double te=-servo_get_position(Left_rear_legJ1);
  double sig=-servo_get_position(Left_rear_legJ3);
  
  //printf("\n b : %f\n",b);
  b=b-beta_init;
  sig=sig+sig_init;
  
//sin(a)*L1
  //printf("\nlegs : alpha : %f beta : %f theta : %f sigma : %f",a,b,te,sig);
  //printf("\nP0 : (%f,%f,%f)\n",w4/2,sin(b+a)*L2,cos(a+b)*L2);
 // printf("\nP2 : (%f,%f,%f)\n",w4/2,-w3+sin(te+teta_init)*L1+sin(sig+te)*L3,-w1-cos(te+teta_init)*L1-cos(te+sig)*L3);
  //printf("\nP2 : (%f,%f,%f)\n",w4/2,-w3+sin(te+teta_init)*L1+sin(sig+te)*L3,-w1-cos(te+teta_init)*L1-cos(sig+te)*L3);
  
  
  
  P0->lignes=4;
  P0->colonnes=1;
  P0->numbers[0][0]=w4/2;
	P0->numbers[1][0]=-w2+sin(a+alpha_init)*L1+sin(b+a)*L2;
  P0->numbers[2][0]=-w1-cos(a+alpha_init)*L1-cos(a+b)*L2;
  P0->numbers[3][0]=1;
  
  P1->lignes=4;
  P1->colonnes=1;
  P1->numbers[0][0]=-w4/2;
	P1->numbers[1][0]=-w2+sin(a+alpha_init)*L1+sin(b+a)*L2;
  P1->numbers[2][0]=-w1-cos(a+alpha_init)*L1-cos(a+b)*L2;
  P1->numbers[3][0]=1;
  
  
  P2->lignes=4;
  P2->colonnes=1;
  P2->numbers[0][0]=w4/2;
	P2->numbers[1][0]=-w3+sin(te+teta_init)*L1+sin(sig+te)*L3;
  P2->numbers[2][0]=-w1-cos(te+teta_init)*L1-cos(sig+te)*L3;
  P2->numbers[3][0]=1;
}

/* ====================================================================== */
//tant qu'il n'est pas dans la position voulue il renvoi 0, 1 sinon
int check_aibobody_correct_position(){
//check if Aibo is in the desired initial position
 
  if( (fabs(servo_get_position(Left_front_legJ1)-angle_left_front_J1)<0.1) && 
      (fabs(servo_get_position(Left_front_legJ3)-angle_left_front_J3)<0.1) && 
      
      (fabs(servo_get_position(Right_front_legJ1)-angle_right_front_J1)<0.1) && 
      (fabs(servo_get_position(Right_front_legJ3)-angle_right_front_J3)<0.1) && 
      
      (fabs(servo_get_position(Left_rear_legJ1)-angle_left_back_J1)<0.1) && 
      (fabs(servo_get_position(Left_rear_legJ3)-angle_left_back_J3)<0.1) &&
      
      (fabs(servo_get_position(Right_rear_legJ1)-angle_right_back_J1)<0.1) && 
      (fabs(servo_get_position(Right_rear_legJ3)-angle_right_back_J3)<0.1) &&
      
      (fabs(servo_get_position(neck_tilt2)-omega)<0.01) &&
      (fabs(servo_get_position(neck_pan)-betha)<0.01) &&
      (fabs(servo_get_position(neck_tilt1)-alpha)<0.01)
  
  ){
    return 1;
  }
  
  printf("Position initiale du corps du robot : leg left front J1 : %f, leg left front J3 : %f, leg right front J1 : %f, leg right front J3 : %f,leg left back J1 : %f, leg left back J3 : %f, leg right back J1 : %f, leg right back J3 : %f \n",servo_get_position(Left_front_legJ1),servo_get_position(Left_front_legJ3),servo_get_position(Right_front_legJ1),servo_get_position(Right_front_legJ3),servo_get_position(Left_rear_legJ1),servo_get_position(Left_rear_legJ3),servo_get_position(Right_rear_legJ1),servo_get_position(Right_rear_legJ3));

  printf("Position initiale de la tte du robot %f %f %f",servo_get_position(neck_tilt2),servo_get_position(neck_pan),servo_get_position(neck_tilt1));
  return 0;
}

/* ====================================================================== */

void move_nearest_paw(Matrice mark){
//determiner quel patte
double x,y,z;
double q1,q2,q3;
double const L1=70.0803110724;
double const L2=76.8969440225;


if(mark.numbers[0][0]>0){ //right paw
//mettre dans referentiel de la patte
//adapte xyz pour qu'ils correspondent au formule de guyot
y=mark.numbers[0][0]-62.5;
x=mark.numbers[1][0]+2.5;
z=-mark.numbers[2][0]-19.5;
}else{
//mettre dans referentiel de la patte
//adapte xyz pour qu'ils correspondent au formule de guyot
y=-mark.numbers[0][0]-62.5;
x=mark.numbers[1][0]+2.5;
z=-mark.numbers[2][0]-19.5;
}

//printf("x,y,z : (%f,%f,%f)",x,y,z);

//printf("(x*x)+(y*y)+(z*z)-(L1*L1)-(L2*L2) : (%f)",((x*x)+(y*y)+(z*z)-(L1*L1)-(L2*L2))/(2*L1*L2));

if(((x*x)+(y*y)+(z*z)-(L1*L1)-(L2*L2))/(2*L1*L2)>1){
  //marque inateignable, mais on suppose que l'on permet que aibo bascule legerement et donc on fixe q3
  if(mark.numbers[0][0]>0) q3=angle_right_front_J3; else q3=angle_left_front_J3;
}else{
q3=acos(((x*x)+(y*y)+(z*z)-(L1*L1)-(L2*L2))/(2*L1*L2)); //plus grand que 1 !!
}

q2=asin((y)/(L1+(L2*cos(q3))));
//q1=atan(z/x)+atan(((L1+L2*cos(q3))*cos(q2))/(L2*sin(q3)));

double a=sin(q3)*L2;
double b=-cos(q2)*cos(q3)*L2-cos(q2)*L1;
double d=b/sin(atan(a/b));
q1=asin(z/d)-atan(a/b);
printf("q1,q2,q3 : (%f,%f,%f)",q1,q2,q3);

if(((x*x)+(y*y)+(z*z)-(L1*L1)-(L2*L2))/(2*L1*L2)<3){
  //move it
  if(mark.numbers[0][0]>0){ //right paw
    servo_set_position(Right_front_legJ1,q1);
    servo_set_position(Right_front_legJ2,q2);
    servo_set_position(Right_front_legJ3,q3);
  }else{
    servo_set_position(Left_front_legJ1,q1);
    servo_set_position(Left_front_legJ2,q2);
    servo_set_position(Left_front_legJ3,q3);
  }
}else{
  printf("Inateignable\n");
}

}




/* ====================================================================== */
//Inverse kinematics
void move_paw(double x_scene_tree,double y_scene_tree,double z_scene_tree){
Matrice angles;
  if(x_scene_tree>0){
  //right leg
    Matrice leg_right=tilt1_to_front_leg_right(1000*x_scene_tree,1000*y_scene_tree,1000*z_scene_tree); //tout en mm
    printf("----------------In leg referential %g %g %g ------------\n",leg_right.numbers[0][0],leg_right.numbers[0][1],leg_right.numbers[0][2]);
  
    angles=calculate_angles_cin(-leg_right.numbers[0][2],leg_right.numbers[0][0],leg_right.numbers[0][1]);
    if (angles.lignes != 0){
      printf("ptr q1,q2,q3 : (%f,%f,%f)",angles.numbers[0][0],angles.numbers[0][1],angles.numbers[0][2]);
      servo_set_position(Right_front_legJ2,angles.numbers[0][1]);
      servo_set_position(Right_front_legJ3,angles.numbers[0][2]);
      servo_set_position(Right_front_legJ1,angles.numbers[0][0]);
    }else printf("unreacheable");
  }
  else{
  //left paw
    Matrice leg_left=tilt1_to_front_leg_left(1000*x_scene_tree,1000*y_scene_tree,1000*z_scene_tree); //tout en mm
    printf("----------------In leg referential %g %g %g ------------\n",leg_left.numbers[0][0],leg_left.numbers[0][1],leg_left.numbers[0][2]);
    
    angles=calculate_angles_cin(-leg_left.numbers[0][2],leg_left.numbers[0][0],leg_left.numbers[0][1]);
        if (angles.lignes != 0){
      printf("ptr q1,q2,q3 : (%f,%f,%f)",angles.numbers[0][0],angles.numbers[0][1],angles.numbers[0][2]);
      servo_set_position(Left_front_legJ2,angles.numbers[0][1]);
      servo_set_position(Left_front_legJ3,angles.numbers[0][2]);
      servo_set_position(Left_front_legJ1,angles.numbers[0][0]);
    }else printf("unreacheable");
  }
}


Matrice calculate_angles_cin(double x, double y,double z){
  Matrice result;
  result.lignes=0;
  result.colonnes=0;
  double q1,q2;
  double q3=0;
  double L1=70.0803110724;
  double L2=76.8969440225;
  double new_z=z;
  
  printf("Position to reach x,y,z : (%f,%f,%f)",x,y,z);
  if(((x*x)+(y*y)+(new_z*new_z)-(L1*L1)-(L2*L2))/(2*L1*L2)>1){
    printf("unreachable");
    L1=L1+20;
    L2=L2+20;
    q3=0.52359;
    new_z=new_z+60;
  }
 if(((x*x)+(y*y)+(new_z*new_z)-(L1*L1)-(L2*L2))/(2*L1*L2)>1){
    printf("really unreachable eaven with prolongation");
  }else{
    result.lignes=1;
    result.colonnes=3;
    if(q3==0) q3=acos(((x*x)+(y*y)+(new_z*new_z)-(L1*L1)-(L2*L2))/(2*L1*L2));
    q2=asin((y)/(L1+(L2*cos(q3))));
    q1=atan(new_z/x)+atan(((L1+L2*cos(q3))*cos(q2))/(L2*sin(q3)));

    printf("q1,q2,q3 : (%f,%f,%f)",q1,q2,q3);
    printf("With robot geometry q1,q2,q3 : (%f,%f,%f)",q1-atan(9/69.5),q2,q3+0.248);

    printf("Verification direct kinematics x %f y %f z %f",
    L2 * cos(q1) * sin(q3) + L2 * sin(q1) * cos(q2) * cos(q3) + L1 * sin(q1) * cos(q2),
    L1 * sin(q2) + L2 * sin(q2) * cos(q3),
    L2 * sin(q1) * sin(q3) - L2 * cos(q1) * cos(q2) * cos(q3) - L1 * cos(q1) * cos(q2));
    result.numbers[0][0]=q1-atan(9/69.5);
    result.numbers[0][1]=q2;
    result.numbers[0][2]=q3+0.248;
  }
  return result;
}


Matrice tilt1_to_front_leg_right(double x,double y,double z){
  Matrice result;
  result.lignes=1;
  result.colonnes=3;
  
  result.numbers[0][0]=x-62.5;
  result.numbers[0][1]=y+19.5;
  result.numbers[0][2]=z-2.5;
  return result;
}

Matrice tilt1_to_front_leg_left(double x,double y,double z){
  Matrice result;
  result.lignes=1;
  result.colonnes=3;
  
  result.numbers[0][0]=-x-62.5;
  result.numbers[0][1]=y+19.5;
  result.numbers[0][2]=z-2.5;
  return result;
}
