/*
 * File:         ers7_jen.c
 * Date:         Januar 2008
 * Description:  A supervisor wich move a mark and compares the real position of the mark with the position calculated by the function calculate_mark_coordinate() 
 * Author:       Jennifer Meinen
 */
#include <device/robot.h>
#include <device/servo.h>
#include <device/camera.h>
#include <math.h>
#include <device/supervisor.h>


#define SIMULATION_STEP 32  // control step duration in milliseconds

// Specifications of the camera
#define FOCAL_LENGTH 3.27 //in mm
#define HORIZ_ANGLE_VIEW 56.9 //degree
#define VERT_ANGLE_VIEW 45.2 //degree
#define horizPictElements 416
#define verticPictElements 320 //The number of picture elements: 416(H) x 320(V)

//Dimensions of aibo head
#define NECK 80
#define HEAD_Y 81.06
#define HEAD_Z 14.6

// For matrice calculation
#define N 4 
typedef struct {
   double numbers[N][N];
   unsigned int lignes;
   unsigned int colonnes;
} Matrice;


/* prototypes */
void  affiche_matrice(Matrice);
Matrice mult(Matrice a, Matrice b);
Matrice inverse3x3(Matrice a); //marche seulement avec des matrices de taille 3x3
Matrice convert_refBody_to_refCam(Matrice Pbody);
Matrice convert_refCam_to_refBody(Matrice Pcam);
int intersect_plane_line(Matrice * point_result,Matrice La,Matrice Lb,Matrice P0,Matrice P1,Matrice P2);
Matrice convert_inhomogenous(Matrice M);
double convert_radian(double degre);
void calcul_middle_parallelogram(double A[2],double B[2],double C[2],double D[2],double result[2]);
int get_mark_pixel(double pixel[2]);
void calculate_pixel_distance(double pixelPosition[2],double pixelDistance[3]);
int isRed_or_Magenta(int r,int g,int b);
void set_initial_position(double j1_left_f,double j3_left_f,double j1_right_f,double j3_right_f,double j1_left_b,double j3_left_b,double j1_right_b,double j3_right_b);
void calculate_P0_P1_P2(Matrice * P0,Matrice * P1,Matrice * P2);
int check_aibobody_correct_position();
Matrice calculate_mark_coordinate();
int mark_in_good_position();
void init_enable_all_devices();
void place_mark_once(double x,double y,double z);
void move_the_mark(double x_init,double x_end,double y_end,int scale);

/*variable globales*/
static int first_time;

//Position of the head to reach
static double position_tete[3]={-1.2,0,0.2};  //neck_tilt1[0] neck_pan[1] neck_tilt2[2]

//Position of the legs to reach
static double angle_left_front_J1=0;
static double angle_left_front_J3=0.5235998; //30 degree in radius
static double angle_right_front_J1=0;
static double angle_right_front_J3=0.5235998;
static double angle_left_back_J1=0;
static double angle_left_back_J3=0.5235998;
static double angle_right_back_J1=0;
static double angle_right_back_J3=0.5235998;

//TO REMEMBER here are the position that were use to produce graphics  
//BAS
//tete en bas {-1.2,0,0.2} 
//tete en bas droite {-1.2,-1,0.2}
//tete gauche {-1.2,1,0.2}

//MIDDLE
//{-0.8,0,0.1}
//{-0.8,1,0.1} tete gauche
////{-0.8,-1,0.1} tete droite

//loin
//{-0.2,0,0.1}

//Results are put in tables to be printed at the end
static double matlab_x[5000];
static double matlab_y[5000];
static double incertitude[5000];
static int matlab_index=0;


//DEVICES
static DeviceTag camera;
//head
static DeviceTag neck_tilt1;
static DeviceTag neck_tilt2;
static DeviceTag neck_pan;
//legs
static DeviceTag Left_front_legJ1;
static DeviceTag Left_front_legJ2;
static DeviceTag Left_front_legJ3;
static DeviceTag Left_rear_legJ1;
static DeviceTag Left_rear_legJ2;
static DeviceTag Left_rear_legJ3;
static DeviceTag Right_front_legJ1 ;
static DeviceTag Right_front_legJ2;
static DeviceTag Right_front_legJ3;
static DeviceTag Right_rear_legJ1;
static DeviceTag Right_rear_legJ2;
static DeviceTag Right_rear_legJ3;

//Tolerance for the color
static double Color_tolerance = 0.25; 

//Supervisor
static NodeRef mark;
static float mark_initial_position[3] = {-0.1, -0.158, -0.1}; 
static float mark_real_position[3] = {0,0,0}; //we will use it to get the value of the mark

static int nb_loops=0;
static int end=0;


static void init(void)
{
  //Enable all devices
  init_enable_all_devices();
  
  //Set the initial position of the head
  servo_set_position (neck_tilt1,position_tete[0]);
  servo_set_position (neck_pan,position_tete[1]);
  servo_set_position (neck_tilt2,position_tete[2]);
  //Set the initial positon of the legs
  set_initial_position(angle_left_front_J1,angle_left_front_J3,angle_right_front_J1,angle_right_front_J3,angle_left_back_J1,angle_left_back_J3,angle_right_back_J1,angle_right_back_J3);
  
  first_time=1;
  
  //Supervisor place the mark in its initial position
  mark=supervisor_node_get_from_def("RedMark"); 
  robot_step(SIMULATION_STEP);
  supervisor_field_set(mark, SUPERVISOR_FIELD_TRANSLATION_X |SUPERVISOR_FIELD_TRANSLATION_Y | SUPERVISOR_FIELD_TRANSLATION_Z, mark_initial_position);
  //mark_real_position will contain the position of the mark updated every SIMULATION_STEP
  supervisor_field_get(mark, SUPERVISOR_FIELD_TRANSLATION_X | SUPERVISOR_FIELD_TRANSLATION_Y | SUPERVISOR_FIELD_TRANSLATION_Z , mark_real_position, SIMULATION_STEP);

    return;
}


static int run(int ms)
{

  if(first_time){
    robot_console_printf("------------Head position : neck_tilt1 = %f neck_pan = %f neck_tilt2 = %f \n",servo_get_position(neck_tilt1),servo_get_position(neck_pan),servo_get_position(neck_tilt2));
    if(check_aibobody_correct_position()){
      first_time=0;
    }
  }
  
  if(nb_loops>=0 && check_aibobody_correct_position() && mark_in_good_position()){
    Matrice result=calculate_mark_coordinate();
    if(result.lignes!=0){
      //We print the position and adapt the units and coordinate xyz because they are not placed the same in the scene tree as in our referential
      robot_console_printf("Position de la marque : x= %.3f , y= %.3f , z= %.3f | Position calculee : x= %.4g , y= %.4g , z= %.4g",1000*mark_real_position[0],
                              -1000*mark_real_position[2],1000*mark_real_position[1],result.numbers[0][0],result.numbers[1][0],result.numbers[2][0]);
      robot_console_printf("Erreur entre position calculee et veritable position : x= %.3f , y= %.3f , z= %.3f\n",fabs((1000*mark_real_position[0])-result.numbers[0][0]),
                              fabs((-1000*mark_real_position[2])-result.numbers[1][0]),fabs((1000*mark_real_position[1])-result.numbers[2][0]));
      
      matlab_x[matlab_index]=1000*mark_real_position[0];
      matlab_y[matlab_index]=-1000*mark_real_position[2];
      incertitude[matlab_index]=sqrt(pow(fabs((1000*mark_real_position[0])-result.numbers[0][0]),2) +  pow(fabs((-1000*mark_real_position[2])-result.numbers[1][0]),2) + pow(fabs((1000*mark_real_position[1])-result.numbers[2][0]),2));
      matlab_index++;
    }
  
    move_the_mark(-0.2,0.2,-0.3,1); //x_start x_end y_end scale :more scale increase less results we have
  
      //TO REMEMBER here are the position that were use to produce graphics  -> square adapted (not to wait to long)
       //NEAR
       //tete en bas {-1.2,0,0.2} -> move_the_mark(-0.2,0.2,-0.3,1);
       //tete en bas droite {-1.2,-1,0.2} move_the_mark(0,0.4,-0.4,1);
       //tete gauche {-1.2,1,0.2}

       //MIDDLE
         //{-0.8,0,0.1}
         //{-0.8,1,0.1} tete gauche
         //{-0.8,-1,0.1} tete droite

       //Far away
        //{-0.2,0,0.1}

    robot_step(SIMULATION_STEP);
    supervisor_field_set(mark, SUPERVISOR_FIELD_TRANSLATION_X |SUPERVISOR_FIELD_TRANSLATION_Y | SUPERVISOR_FIELD_TRANSLATION_Z, mark_initial_position);
    nb_loops++;
  }
  
  if(nb_loops<0 && !end){
    //End of simulation result for matlab must be printed 
    printf("Vecteur x = [");
    int i;
    for(i=0;i<matlab_index;i++){
      printf("%.2f \t",matlab_x[i]);
      fflush( stdin );
      if(i<matlab_index-1) printf(";");
    }
    printf("]\n");
    
    printf("Vecteur y = [");
    for(i=0;i<matlab_index;i++){
      printf("%.2f \t",matlab_y[i]);
      fflush( stdin );
      if(i<matlab_index-1) printf(";");
    }
    printf("]\n");    

    printf("Vecteur incertitude = [");
    for(i=0;i<matlab_index;i++){
      printf("%.2f\t",incertitude[i]);
      if(i<matlab_index-1) printf(";");
    }
    printf("]\n");
    robot_console_printf("-------------------END SIMULATION");
    end=1;
  }




  /*
  //For a test with only one position
  place_mark_once(-0.1, -0.158, -0.1);
  */

    return SIMULATION_STEP;
}

static void die(void)
{
    return;
}

int main(void)
{
    robot_live(init);
    robot_die(die);
    robot_run(run);
    return 0;
}

//Place the mark at a new position each time called
//y_start end the step of y and z will not be modified because they need to be ajusted on the initial position
void move_the_mark(double x_init,double x_end,double y_end,int scale){ 
//evolve the long of a square
   if(mark_initial_position[0]>x_end)
          {
            mark_initial_position[0]=x_init;
            if(mark_initial_position[2]<y_end)
            {
              nb_loops=-43;
            }
            else
            {
              mark_initial_position[2]-=0.02*scale;
              mark_initial_position[1]+=0.00064*scale;
            }
          }
          else
          {
              mark_initial_position[0]+=0.005*scale;
          }
}

//Place the mark at one position and calculate the position
void place_mark_once(double x,double y,double z){
  if(nb_loops<=1 && check_aibobody_correct_position()){
    Matrice result=calculate_mark_coordinate();
    if(result.lignes!=0){
    robot_console_printf("------------Position de la tete : neck_tilt1 = %f neck_pan = %f neck_tilt2 = %f \n",servo_get_position(neck_tilt1),servo_get_position(neck_pan),servo_get_position(neck_tilt2));
    robot_console_printf("Position de la marque : x= %.3f , y= %.3f , z= %.3f | Position calculee : x= %.4g , y= %.4g , z= %.4g",1000*mark_real_position[0],
                              -1000*mark_real_position[2],1000*mark_real_position[1],result.numbers[0][0],result.numbers[1][0],result.numbers[2][0]);
    robot_console_printf("Erreur entre position calculee et veritable position : x= %.3f , y= %.3f , z= %.3f\n",fabs((1000*mark_real_position[0])-result.numbers[0][0]),
                              fabs((-1000*mark_real_position[2])-result.numbers[1][0]),fabs((1000*mark_real_position[1])-result.numbers[2][0]));
  }else{
    robot_console_printf("Mark not visible");
  }
    mark_initial_position[0]=x;
    mark_initial_position[1]=y;
    mark_initial_position[2]=z;
    supervisor_field_set(mark, SUPERVISOR_FIELD_TRANSLATION_X |SUPERVISOR_FIELD_TRANSLATION_Y | SUPERVISOR_FIELD_TRANSLATION_Z, mark_initial_position);
    nb_loops++;
  }
}
/* ====================================================================== */
//Called once in the initialization it initialyze the camera and all the servo of the legs to be able to read values
void init_enable_all_devices(){
  //camera
  camera = robot_get_device("PRM:/r1/c1/c2/c3/i1-FbkImageSensor:F1");
  camera_enable(camera, SIMULATION_STEP);
 
  // head 
  neck_tilt1 = robot_get_device("PRM:/r1/c1-Joint2:11");
  servo_enable_position  (neck_tilt1, SIMULATION_STEP);
  
  neck_pan = robot_get_device("PRM:/r1/c1/c2-Joint2:12");
  servo_enable_position  (neck_pan, SIMULATION_STEP);
  
  neck_tilt2 = robot_get_device("PRM:/r1/c1/c2/c3-Joint2:13");
  servo_enable_position  (neck_tilt2, SIMULATION_STEP);
  
  // legs
  Left_front_legJ1 = robot_get_device("PRM:/r2/c1-Joint2:21");
  Left_front_legJ2 = robot_get_device("PRM:/r2/c1/c2-Joint2:22");
  Left_front_legJ3 = robot_get_device("PRM:/r2/c1/c2/c3-Joint2:23");
  
  Left_rear_legJ1 = robot_get_device("PRM:/r3/c1-Joint2:31");
  Left_rear_legJ2 = robot_get_device("PRM:/r3/c1/c2-Joint2:32");
  Left_rear_legJ3 = robot_get_device("PRM:/r3/c1/c2/c3-Joint2:33");
  
  Right_front_legJ1 = robot_get_device("PRM:/r4/c1-Joint2:41");
  Right_front_legJ2 = robot_get_device("PRM:/r4/c1/c2-Joint2:42");
  Right_front_legJ3 = robot_get_device("PRM:/r4/c1/c2/c3-Joint2:43");
  
  Right_rear_legJ1 = robot_get_device("PRM:/r5/c1-Joint2:51");
  Right_rear_legJ2 = robot_get_device("PRM:/r5/c1/c2-Joint2:52");
  Right_rear_legJ3 = robot_get_device("PRM:/r5/c1/c2/c3-Joint2:53");
  
  servo_enable_position  (Left_front_legJ1, SIMULATION_STEP);
  servo_enable_position  (Left_front_legJ2, SIMULATION_STEP);
  servo_enable_position  (Left_front_legJ3, SIMULATION_STEP);
  
  servo_enable_position  (Left_rear_legJ1, SIMULATION_STEP);
  servo_enable_position  (Left_rear_legJ2, SIMULATION_STEP);
  servo_enable_position  (Left_rear_legJ3, SIMULATION_STEP);
  
  servo_enable_position  (Right_front_legJ1, SIMULATION_STEP);
  servo_enable_position  (Right_front_legJ2, SIMULATION_STEP);
  servo_enable_position  (Right_front_legJ3, SIMULATION_STEP);
  
  servo_enable_position  (Right_rear_legJ1, SIMULATION_STEP);
  servo_enable_position  (Right_rear_legJ2, SIMULATION_STEP);
  servo_enable_position  (Right_rear_legJ3, SIMULATION_STEP);

}

/* ====================================================================== */
//This function is needed only for the supervisor it checks that the mark is correctly placed in the desired position
int mark_in_good_position(){
  if(fabs(mark_initial_position[0]-mark_real_position[0])<0.00000001
  && fabs(mark_initial_position[1]-mark_real_position[1])<0.00000001
  && fabs(mark_initial_position[2]-mark_real_position[2])<0.00000001
  ){
    return 1;
  }else{
  robot_console_printf("Mark is not yet placed");
  supervisor_field_set(mark, SUPERVISOR_FIELD_TRANSLATION_X |SUPERVISOR_FIELD_TRANSLATION_Y | SUPERVISOR_FIELD_TRANSLATION_Z, mark_initial_position);
    return 0;
  }
}

/* ====================================================================== */
//Calculates the position of the mark
Matrice calculate_mark_coordinate(){
  Matrice O = {{
               {0},
               {0},
               {0} } , 3 , 1 };
 double pixelPosition[2];
 double pixelDistance[3];
 Matrice G1;
 Matrice G2;
 Matrice G3;
 calculate_P0_P1_P2(&G1,&G2,&G3);
      
 Matrice mark;
 mark.lignes=0;
 mark.colonnes=0;         
 
 if(get_mark_pixel(pixelPosition)){
   Matrice G1_cam = convert_refBody_to_refCam(G1);
   Matrice G2_cam = convert_refBody_to_refCam(G2);
   Matrice G3_cam = convert_refBody_to_refCam(G3);
        
   //robot_console_printf("G1 dans referentiel de la camera : ");
   //affiche_matrice (G1_cam);
   //robot_console_printf("\n");
        
   calculate_pixel_distance(pixelPosition,pixelDistance);
        
   Matrice Lb = { {
                  {pixelDistance[0]},
                  {pixelDistance[1]},
                  {pixelDistance[2]} } , 3 , 1 };
   //robot_console_printf("Represente la distance reel d'un pixel : ");
   //affiche_matrice (Lb);
   //robot_console_printf("\n");
        
   Matrice point_result;
        
   if( intersect_plane_line(&point_result,O,Lb,convert_inhomogenous(G1_cam),convert_inhomogenous(G2_cam),convert_inhomogenous(G3_cam))  ) {
     //un point a été trouvé : point_result est homogène on peut donc appliquer directement la conversion dans l'autre referentiel et affiché le résultat dans la console
     //robot_console_printf("Point dans le referentiel de la camera : ");
     //affiche_matrice (point_result);
     //robot_console_printf("\n");
     //robot_console_printf("Position de la marque calculée : ");
     mark=convert_refCam_to_refBody(point_result);
     //affiche_matrice (mark);
     //robot_console_printf("\n");
    }else{
	  robot_console_printf("Aucune intersection avec le plan n a ete trouvee");
    }
  }else{
    //robot_console_printf("Pas de pixel rouge");
  }

return mark;
}
/* ====================================================================== */

//This function calculates the real physic distance that corresponds to the position of a pixel in the image. It takes in account the focal length and the angle of view of the camera
void calculate_pixel_distance(double pixelPosition[2],double pixelDistance[3]){
 float image_width = camera_get_width  (camera);
 float image_height = camera_get_height (camera); 
 
 float h = image_width/2;
 float v = image_height/2;
 double a = camera_get_fov(camera)/2.0;
 double b = convert_radian(VERT_ANGLE_VIEW/2.0); 
 //robot_console_printf("angle a: %f et b:%f",a,b);
 //robot_console_printf("Pixel depuis le bord (%f,%f)\n",pixelPosition[0],pixelPosition[1]);
 //robot_console_printf("Pixel dans le referentiel x,y milieu de l'image (%f,%f)\n",pixelPosition[0]-h,v-pixelPosition[1]);
 
 pixelDistance[0]= ((pixelPosition[0]-h)* FOCAL_LENGTH * tan(a))/h;
 pixelDistance[1]= FOCAL_LENGTH;
 pixelDistance[2]= ((v-pixelPosition[1])* FOCAL_LENGTH * tan(b))/v;
}


/* ====================================================================== */

//Takes a picture checks if a red mark exist , if yes returns 1 and put the result of the mark inside pixel
int get_mark_pixel(double pixel[2]){
  const unsigned char *image = camera_get_image(camera);
  unsigned short image_width = camera_get_width  (camera);
  unsigned short image_height = camera_get_height (camera); 
  int x,y;
  
  double A[2], B[2], C[2], D[2];
  
  double A_temp[2], B_temp[2];
  
  int nb_red_ligne_prec=0;
  int nb_red_ligne_curr=0;
  int has_red=0;

  for (y = 0; y < image_height; y++){ //line of image
    for (x = 0; x < image_width; x++) { //colonn of image
      int r = camera_image_get_red(image, image_width, x, y);
      int g = camera_image_get_green(image, image_width, x, y);
      int b = camera_image_get_blue(image, image_width, x, y);
      
      if(isRed_or_Magenta(r,g,b)){
        if(nb_red_ligne_prec==0){
          //We first stepped on the mark
          if(nb_red_ligne_curr==0){
            //First pixel of the line
            has_red=1;
            A[0]=x;
            A[1]=y;
            C[0]=x;
            C[1]=y;
            B[0]=x;
            B[1]=y;
            D[0]=x;
            D[1]=y;          
            nb_red_ligne_curr=1;
          }else{
            B[0]=x;
            B[1]=y;
            D[0]=x;
            D[1]=y;
            nb_red_ligne_curr++;
          }
          
        }else{
          //last line already had red pixels
          if(nb_red_ligne_curr==0){
              //A must change only if the number of pixel of last line was less than this one
              A_temp[0]=x;
              A_temp[1]=y;
              nb_red_ligne_curr=1;
            }else{
              B_temp[0]=x;
              B_temp[1]=y;
              nb_red_ligne_curr++;
            }
        }
      }
      
    }
    if(!(nb_red_ligne_prec==0) && nb_red_ligne_prec<nb_red_ligne_curr){ 
      A[0]=A_temp[0];
      A[1]=A_temp[1];
      B[0]=B_temp[0];
      B[1]=B_temp[1];
    }
    
    if(!(nb_red_ligne_prec==0) && nb_red_ligne_prec==nb_red_ligne_curr){ 
      C[0]=A_temp[0];
      C[1]=A_temp[1];
      D[0]=B_temp[0];
      D[1]=B_temp[1];
    }
    
    //all pixels of a line have been visited
    nb_red_ligne_prec=nb_red_ligne_curr;
    nb_red_ligne_curr=0;
  }

  if(has_red){
    calcul_middle_parallelogram(A,B,C,D,pixel);
    return 1;
  }else{
    return 0;
  }
}

/* ====================================================================== */
//Returns 1 if pixel is red or magenta
int isRed_or_Magenta(int r,int g,int b){
  if(  r>(255-(Color_tolerance*255)) && g<100 ){
    return 1;
  }else{
    return 0;
  }
}

//Returns 1 if pixel is red
int isRed(int r,int g,int b){
  if(r>(255-Color_tolerance*255) && g<Color_tolerance*255 && b<Color_tolerance*255){
    return 1;
  }else{
    return 0;
  }
}


/* ====================================================================== */
//Multiply two matrices and return the result
Matrice mult(Matrice a, Matrice b)
{
  Matrice resultat;
  unsigned int i, j, k;

  resultat.lignes = a.lignes;
  resultat.colonnes = b.colonnes;
  
  if (a.colonnes == b.lignes) {
    
    for (i = 0; i < a.lignes; ++i) {       
      for (j = 0; j < b.colonnes; ++j) {   
        resultat.numbers[i][j] = 0.0;
	  for (k = 0; k < b.lignes; ++k) { 
	  resultat.numbers[i][j] += a.numbers[i][k] * b.numbers[k][j];
        }
      }
    }
  }

  return resultat;
}


/* ====================================================================== */
//Prints a matrice on the console
void affiche_matrice(Matrice m)
{
  unsigned int i, j;
  if(m.lignes==4 && m.colonnes==1)
  {
    robot_console_printf("(%g,%g,%g) \n", m.numbers[0][0],m.numbers[1][0],m.numbers[2][0]);
  }
  else
  {
    for (i = 0; i < m.lignes; ++i) {
      for (j = 0; j < m.colonnes; ++j) 
        robot_console_printf("%g ", m.numbers[i][j]);
      putchar('\n');
    }
  }
}



/* ====================================================================== */
//Inverse matrice only for a 3x3 matrice
Matrice inverse3x3(Matrice A)
{
  Matrice resultat;
  resultat.lignes=0;
  resultat.colonnes=0;
  
  if(A.lignes == A.colonnes && A.lignes == 3){
    double a=A.numbers[0][0], b=A.numbers[0][1], c=A.numbers[0][2], 
           d=A.numbers[1][0], e=A.numbers[1][1], f=A.numbers[1][2],
           g=A.numbers[2][0], h=A.numbers[2][1], i=A.numbers[2][2];
    double detA = a*(e*i-f*h) - b*(d*i-f*g) + c*(d*h-e*g);
	
    if(fabs(detA)>0){
      resultat.numbers[0][0]=(e*i-f*h)/detA;
      resultat.numbers[0][1]=(c*h-b*i)/detA;
      resultat.numbers[0][2]=(b*f-c*e)/detA;
      resultat.numbers[1][0]=(f*g-d*i)/detA;
      resultat.numbers[1][1]=(a*i-c*g)/detA;
      resultat.numbers[1][2]=(c*d-a*f)/detA;
      resultat.numbers[2][0]=(d*h-e*g)/detA;
      resultat.numbers[2][1]=(b*g-a*h)/detA;
      resultat.numbers[2][2]=(a*e-b*d)/detA;
      resultat.lignes=3; 
      resultat.colonnes=3;
     } 
  }
  return resultat;
}

/* ====================================================================== */
Matrice rotateX(double tetha){
  Matrice aroundX = {  {{1,0,0,0},
                        {0,cos(tetha),sin(tetha),0},
                        {0,-sin(tetha),cos(tetha),0},
                        {0,0,0,1} } , 4 , 4 };
  return aroundX;
}

Matrice rotateZ(double tetha){
  Matrice aroundZ = {  {{cos(tetha),sin(tetha),0,0},
                      {-sin(tetha),cos(tetha),0,0},
                      {0,0,1,0},
                      {0,0,0,1} } , 4 , 4 };
  return aroundZ;
}

Matrice translate(double a,double b,double c){
  Matrice translation = {  {{1,0,0,-a},
                        {0,1,0,-b},
                        {0,0,1,-c},
                        {0,0,0,1} } , 4 , 4 };
  return translation;
}

//Convert a point from the referential of the body to the referential of the camera
Matrice convert_refBody_to_refCam(Matrice Pbody){
  Matrice aroundX1 = rotateX(servo_get_position(neck_tilt1));
  Matrice aroundZ = rotateZ(servo_get_position(neck_pan));
  Matrice trans1 = translate(0,0,NECK);
  Matrice aroundX2 = rotateX(servo_get_position(neck_tilt2));
  Matrice trans2 = translate(0,HEAD_Y,-HEAD_Z);
  return mult(trans2 , mult(aroundX2 , mult(trans1, mult(aroundZ, mult(aroundX1,Pbody) ) ) ) );
}

//Convert a point from the referential of the camera to the referential of the body
Matrice convert_refCam_to_refBody(Matrice Pcam){
  Matrice aroundX1 = rotateX(-servo_get_position(neck_tilt1));
  Matrice aroundZ = rotateZ(-servo_get_position(neck_pan));
  Matrice trans1 = translate(0,0,-NECK);
  Matrice aroundX2 = rotateX(-servo_get_position(neck_tilt2));
  Matrice trans2 = translate(0,-HEAD_Y,HEAD_Z);
  return mult(aroundX1 , mult(aroundZ,mult(trans1,mult(aroundX2,mult(trans2,Pcam)))));
}

/* ====================================================================== */

//Calculates the intersection between a line and a plane if an intersection is found returns 1 else 0, the result is put inside point_result
int intersect_plane_line(Matrice * point_result,Matrice La,Matrice Lb,Matrice P0,Matrice P1,Matrice P2){
  Matrice M={{ {La.numbers[0][0]-Lb.numbers[0][0],P1.numbers[0][0]-P0.numbers[0][0],P2.numbers[0][0]-P0.numbers[0][0]},
	       {La.numbers[1][0]-Lb.numbers[1][0],P1.numbers[1][0]-P0.numbers[1][0],P2.numbers[1][0]-P0.numbers[1][0]},
	       {La.numbers[2][0]-Lb.numbers[2][0],P1.numbers[2][0]-P0.numbers[2][0],P2.numbers[2][0]-P0.numbers[2][0]}
             }, 3, 3 };
  Matrice V={{ {La.numbers[0][0]-P0.numbers[0][0]},
               {La.numbers[1][0]-P0.numbers[1][0]},
               {La.numbers[2][0]-P0.numbers[2][0]}
              }, 3, 1 };
  int i;
  Matrice Temp=inverse3x3(M);
	
  if(Temp.lignes != 0){
    Temp=mult(Temp,V);
    for(i=0;i<3;i++){
      point_result->numbers[i][0]=La.numbers[i][0]+Temp.numbers[0][0]*(Lb.numbers[i][0]-La.numbers[i][0]);
    }
    point_result->numbers[3][0]=1;
    point_result->lignes=4;
    point_result->colonnes=1;
    return 1;
  }
  else{
    return 0;
  }
}


/* ====================================================================== */
Matrice convert_inhomogenous(Matrice M){
  Matrice result;
  int i;
  if(M.lignes == 4 && M.colonnes == 1) {
    for(i=0;i<M.lignes-1;i++){
      result.numbers[i][0]=M.numbers[i][0]/M.numbers[3][0];
    }
    result.lignes = 3;
    result.colonnes = 1;
  }else if(M.lignes == 1 && M.colonnes == 4){
    for(i=0;i<M.colonnes-1;i++){
      result.numbers[0][i]=M.numbers[0][i]/M.numbers[0][3];
    }
  result.lignes = 1;
  result.colonnes = 3;
  }else {
  result.lignes = 0;
  result.colonnes = 0;
  }
  return result;
}

/* ====================================================================== */
double convert_radian(double degre){
return (degre*2*M_PI)/360.0;
}

/* ====================================================================== */
//Calculates the middle of a parallelogram
void calcul_middle_parallelogram(double A[2],double B[2],double C[2],double D[2],double result[2]){
 double A_car[2];
 double B_car[2];
 double C_car[2];
 if(A[0]<=C[0]){
   A_car[0]=C[0];
   C_car[0]=C[0];
 }
 else{
   A_car[0]=A[0];
   C_car[0]=A[0];
 }
 
 A_car[1]=A[1];
 C_car[1]=C[1];

 if(B[0]<=D[0]){
   B_car[0]=B[0];
 }
 else{
   B_car[0]=D[0];
 }
 B_car[1]=B[1];
 result[0]=A_car[0]+((B_car[0]-A_car[0])/2);
 result[1]=A_car[1]+((C_car[1]-A_car[1])/2);
}

/* ====================================================================== */
//Set initial position of aibo
void set_initial_position(double j1_left_f,double j3_left_f,double j1_right_f,double j3_right_f,double j1_left_b,double j3_left_b,double j1_right_b,double j3_right_b){

  servo_set_position(Left_front_legJ1, j1_left_f);
  servo_set_position(Left_front_legJ2, 0);
  servo_set_position(Left_front_legJ3, j3_left_f);
  
  servo_set_position(Right_front_legJ1, j1_right_f);
  servo_set_position(Right_front_legJ2, 0);
  servo_set_position(Right_front_legJ3, j3_right_f);
  
  servo_set_position(Left_rear_legJ1, j1_left_b);
  servo_set_position(Left_rear_legJ2, 0);
  servo_set_position(Left_rear_legJ3, j3_left_b);
  
  servo_set_position(Right_rear_legJ1, j1_right_b);
  servo_set_position(Right_rear_legJ2, 0);
  servo_set_position(Right_rear_legJ3, j3_right_b);
}


/* ====================================================================== */
//Knowing the position of servos in legs calculates three points on the floor, these points are in fact the point where the pow touches the floor
void calculate_P0_P1_P2(Matrice * P0,Matrice * P1,Matrice * P2){
  double const w1=19.5;
  double const w2=2.5;
  double const w3=132.5;
  double const w4=134.4;
  double const L1=70.0803110724;
  double const L2=76.8969440225;
  double const L3=79.4099489989;
  double const alpha_init=atan(9.0/69.5);
  double const teta_init=-alpha_init;
  double const beta_init=convert_radian(30)-atan(28.3/71.5);
  double const sig_init=convert_radian(30)-atan(21.3/76.5);
  
  //left and right pow move together
  double a=servo_get_position(Left_front_legJ1); 
  double b=servo_get_position(Left_front_legJ3);
      
  double te=-servo_get_position(Left_rear_legJ1);
  double sig=-servo_get_position(Left_rear_legJ3);
  
  b=b-beta_init;
  sig=sig+sig_init;
    
  P0->lignes=4;
  P0->colonnes=1;
  P0->numbers[0][0]=w4/2;
  P0->numbers[1][0]=-w2+sin(a+alpha_init)*L1+sin(b+a)*L2;
  P0->numbers[2][0]=-w1-cos(a+alpha_init)*L1-cos(a+b)*L2;
  P0->numbers[3][0]=1;
  
  P1->lignes=4;
  P1->colonnes=1;
  P1->numbers[0][0]=-w4/2;
  P1->numbers[1][0]=-w2+sin(a+alpha_init)*L1+sin(b+a)*L2;
  P1->numbers[2][0]=-w1-cos(a+alpha_init)*L1-cos(a+b)*L2;
  P1->numbers[3][0]=1;
    
  P2->lignes=4;
  P2->colonnes=1;
  P2->numbers[0][0]=w4/2;
  P2->numbers[1][0]=-w3+sin(te+teta_init)*L1+sin(sig+te)*L3;
  P2->numbers[2][0]=-w1-cos(te+teta_init)*L1-cos(sig+te)*L3;
  P2->numbers[3][0]=1;
}

/* ====================================================================== */
//Only for test in simulation to be sure aibo is in the position we want, not applicable on the real robot because if we specifie a position aibo can have up to 7 degrees of incertitude
int check_aibobody_correct_position(){
  if( 
    (fabs(servo_get_position(Left_front_legJ1)-angle_left_front_J1)<0.001) && 
    (fabs(servo_get_position(Left_front_legJ3)-angle_left_front_J3)<0.001) && 
      
    (fabs(servo_get_position(Right_front_legJ1)-angle_right_front_J1)<0.001) && 
    (fabs(servo_get_position(Right_front_legJ3)-angle_right_front_J3)<0.001) && 
     
    (fabs(servo_get_position(Left_rear_legJ1)-angle_left_back_J1)<0.001) && 
    (fabs(servo_get_position(Left_rear_legJ3)-angle_left_back_J3)<0.001) &&
      
    (fabs(servo_get_position(Right_rear_legJ1)-angle_right_back_J1)<0.001) && 
    (fabs(servo_get_position(Right_rear_legJ3)-angle_right_back_J3)<0.001) &&
      
    (fabs(servo_get_position(neck_tilt2)-position_tete[2])<0.001) &&
    (fabs(servo_get_position(neck_pan)-position_tete[1])<0.001) &&
    (fabs(servo_get_position(neck_tilt1)-position_tete[0])<0.001)
  ){
    return 1;
  }
  else{
      return 0;
  }
}

